#include "stdafx.h"
#include "Monitors.h"

//#define DEBUGTRACE

#ifdef DEBUGTRACE
#define MON_TRACE(x) TRACE x
#else
#define MON_TRACE(x)
#endif


/****************************************************************************
*                            CMonitors::CMonitors
* Effect: 
*       constructor
****************************************************************************/

CMonitors::CMonitors()
    {
     valid = FALSE; // no monitors found at this point
    } // CMonitors::CMonitors

/****************************************************************************
*                           CMonitors::monitorproc
* Inputs:
*       HMONITOR monitor: the monitor handle
*       HDC: ignored
*       LPRECT: ignored
* Result: BOOL
*       TRUE to continue enumeration
*       FALSE to terminate
* Effect: 
*       Adds a MONITORINFOEX element to the monitors array
*       If there is an error, sets valid to FALSE
****************************************************************************/

BOOL CMonitors::monitorproc(HMONITOR monitor, HDC, LPRECT)
    {
     CString s;
     MONITORINFOEX info;
     info.cbSize = sizeof(info);
     if(!::GetMonitorInfo(monitor, (LPMONITORINFO)&info))
        { /* failed */
         monitors.RemoveAll();  // kill off all partial information
         valid = FALSE;
         return FALSE;  // stop enumeration if error
        } /* failed */

     monitors.Add(info);
     return TRUE;
    }

/****************************************************************************
*                           CMonitors::monitorproc
* Inputs:
*       HMONITOR monitor: Monitor handle
*       HDC dc: NULL
*       LPRECT rect: NULL
*       LPARAM data: (LPARAM)(CMonitors *)
* Result: BOOL
*       TRUE to stop enumeration
*       FALSE to continue enumeration
* Effect: 
*       Calls the internal enumerator
****************************************************************************/

/* static */ BOOL CALLBACK CMonitors::monitorproc(HMONITOR monitor, HDC dc, LPRECT rect, LPARAM data)
   {
    CMonitors * self = (CMonitors *)data;
    return self->monitorproc(monitor, dc, rect);
   }


/****************************************************************************
*                           CMonitors::Initialize
* Result: BOOL
*       TRUE if successful
*       FALSE if error (::GetLastError reveals reason)
* Effect: 
*       Creates an array of MONITORINFOEX objects for the monitors which are
*       on the system
* Notes:
*       This should be called whenever there is a WM_DISPLAYCHANGE event
****************************************************************************/

BOOL CMonitors::Initialize()
    {
     valid = TRUE; // assume true, any error will set it false

     monitors.RemoveAll();
     SortX.RemoveAll();
     SortY.RemoveAll();
     VirtualRects.RemoveAll();

     if(!::EnumDisplayMonitors(NULL, NULL, monitorproc, (LPARAM)this))
        { /* failed */
         valid = FALSE;
         return FALSE;
        } /* failed */

     if(valid)
        { /* sort them */
         Sort(SortX, ByX);
         Sort(SortY, ByY);
        } /* sort them */

     // Now fill in the rectangles
     VirtualRects.SetSize(monitors.GetSize());

     for(int i = 0; i < monitors.GetSize(); i++)
        { /* fix up virtual */
         VirtualRects[i] = monitors[i].rcMonitor;
         ScreenToVirtual(VirtualRects[i]);
        } /* fix up virtual */
     return valid;
    } // CMonitors::Initialize

/****************************************************************************
*                               CMonitors::Sort
* Inputs:
*       CArray<int, int> & data: Data vector to sort
*       SortBy what: What to sort by
* Result: void
*       
* Effect: 
*       Sorts the list.  Note this uses Bubble Sort, which you would think
*       would be bad until you realize that it is very rare that n is ever
*       greater than 2.  And it probably isn't much larger than 4 even in
*       extreme cases
****************************************************************************/

void CMonitors::Sort(CArray<int, int> & data, SortBy what)
    {
     if(data.GetSize() == 0)
        { /* initialize */
         data.SetSize(monitors.GetSize());
         for(int i = 0; i < data.GetSize(); i++)
            data[i] = i;
        } /* initialize */

     if(data.GetSize() < 2)
        return; // size 0 or 1 is already sorted
     
     BOOL changed = TRUE;
     while(changed)
        { /* sort pass */
         changed = FALSE;  // assume no changes, it is sorted
         for(int i = 0; i < data.GetSize() - 1; i++)
            { /* check elements */
             BOOL doswap = FALSE;
             switch(what)
                { /* what */
                 case ByX:
                    doswap = monitors[data[i]].rcMonitor.left > monitors[data[i+1]].rcMonitor.left;
                    break;
                 case ByY:
                    doswap = monitors[data[i]].rcMonitor.top > monitors[data[i+1]].rcMonitor.top;
                    break;
                } /* what */
             if(doswap)
                { /* swap it */
                 int t = data[i];
                 data[i] = data[i + 1];
                 data[i + 1] = t;
                 changed = TRUE;
                } /* swap it */
            } /* check elements */
        } /* sort pass */
    } // CMonitors::Sort

/****************************************************************************
*                         CMonitors::ScreenToVirtual
* Inputs:
*       CRect & r: IN: Rectangle which represents screen coordinates
*                 OUT: Rectangle which represents virtual coordinates
* Result: BOOL
*       TRUE if conversion was successful
*       FALSE if there was an error
* Effect: 
*       Converts the coordinates of the rectangle to virtual coordinates
****************************************************************************/

BOOL CMonitors::ScreenToVirtual(CRect & r)
    {
     if(!ScreenToVirtual((CPoint &)r.left))
        return FALSE;
     if(!ScreenToVirtual((CPoint &)r.right))
        return FALSE;
     return TRUE;
    } // CMonitors::ScreenToVirtual

/****************************************************************************
*                            CMonitors::CheckValid
* Result: BOOL
*       TRUE if first-order validation tests pass
*       FALSE if first-order validation tests fail
* Effect:
*       If there is a validity problem, calls SetLastError() to set the
*       erorr code
****************************************************************************/

BOOL CMonitors::CheckValid()
    {
     if(!valid)
        { /* uninitialized */
         ::SetLastError(MONITORS_ERR_UNINITIALIZED); // not initialized
         return FALSE;
        } /* uninitialized */

     if(monitors.GetSize() == 0)
        { /* initialization failure */
         ::SetLastError(MONITORS_ERR_INIT_FAILURE); // this should be impossible
         return FALSE;
        } /* initialization failure */
     return TRUE;
    } // CMonitors::CheckValid

/***********************************************************************************************************
*                         CMonitors::ScreenToVirtual
* Inputs:
*       CPoint & pt: IN: Point in screen coordinates
*                   OUT: Point in virtual coordinates
* Result: BOOL
*       TRUE if successful
*       FALSE if error
*               Coordinate is outside virtual screen (ERROR_DS_SIZELIMIT_EXCEEDED)
*               Initialization not done  (OLE_E_BLANK)
*               Initialization failed    (ERROR_INTERNAL_ERROR)
* Effect: 
*       Converts the point
* Notes:
*************************************************************************************************************
*                               B Screen(0,-2) Virtual(24,0)
*                               C Screen(0,0)  Virtual(24,2)
*         Screen(-24,-2)        |                      D Screen(23,1) Virtual(47,3)
*       X Virtual (0,0)         |                      |                   E Screen(43,1) Virtual(67,3)
*       |                       |                      |                   |
*       |   -    -    -         |                      |                   V
*       V   2    1    1    -   -V         1    1    2  V 2    3    3    4  4 \
*       43210987654321098765432101234567890123456789012345678901234567890123  > Screen X coordinates
*       |                       ............................................ /
*       V         1    1    2    2    3    3    4    4    5    5    6    6 6 \
*       01234567890123456789012345678901234567890123456789012345678901234567  > Virtual X coordinates
*       X-----------------------B----------------------+-------------------+ / 0   -2
*       |#######################|######################|###################|   1   -1
*       |#######################C----------------------+###################|   2    0
*  +--->A-----------------------+                      D-------------------E   3    1
*  |    |                       |      *U              |                   |   4    2
*  |    |                       |          2           |        3          |   5    3
*  |    |           1           |                      |                   |   6    4
*  |    |                       |                      |             *V    |   7    5
*  |    |      *T               |                      |                   |   8    6
*  |    |                       |                      +-------------------+   9    7
*  |    |                       +----------------------+###################|  10    8
*  |    |                       |##########################################|  11    9
*  |    +-----------------------+------------------------------------------+  12   10
*  |                                                                       ^   ^    ^
*  |                                                                       |   |    |
*  |                               (SM_CXVIRTUALSCREEN, SM_CYVIRTUALSCRREEN)   |    Screen Y coordinates
*  A Screen(-24, 3) Virtual(0,3)                                               Virtual Y coordinates
*       
*
*       T: Screen(-17, 5) Virtual(7, 8)
*       U: Screen(7, 2)   Virtual(31, 4)
*       V: Screen(37, 5)  Virtual(61, 7)
************************************************************************************************************/

BOOL CMonitors::ScreenToVirtual(CPoint & pt)
    {
     if(!CheckValid())
        return FALSE;

     if(monitors.GetSize() == 1)
        return TRUE;  // already in virtual coordinates...only one monitor

     MON_TRACE(("ScreenToVirtual: pt=(%d,%d)\n", pt.x, pt.y));
     
     CPoint v(0,0);
     //****************************************************************
     // Convert X
     //****************************************************************
     BOOL found = FALSE;
     
     for(int i = 0; i < monitors.GetSize(); i++)
        { /* scan X and accumulate */
         MON_TRACE(("ScreenToVirtual: SortX[%d].rcMonitor = (L %d, T %d, R %d, B %d) [%d, %d]\n", 
               i,
               monitors[SortX[i]].rcMonitor.left,
               monitors[SortX[i]].rcMonitor.top,
               monitors[SortX[i]].rcMonitor.right,
               monitors[SortX[i]].rcMonitor.bottom,
               monitors[SortX[i]].rcMonitor.right - monitors[SortX[i]].rcMonitor.left,
               monitors[SortX[i]].rcMonitor.bottom - monitors[SortX[i]].rcMonitor.top));

         if(pt.x <= monitors[SortX[i]].rcMonitor.right)
            { /* within current */
             v.x = - monitors[SortX[0]].rcMonitor.left + pt.x;
             found = TRUE;
             break; // we've resolved it
            } /* within current */
        } /* scan X and accumulate */

     if(!found)
        { /* out of range */
         ::SetLastError(MONITORS_ERR_OUT_OF_BOUNDS);
         return FALSE;  // point is outside virtual screen
        } /* out of range */

     MON_TRACE(("ScreenToVirtual: v.x=%d\n", v.x));

     //****************************************************************
     // Convert Y
     //****************************************************************
     found = FALSE;
     for(int i = 0; i < monitors.GetSize(); i++)
        { /* scan Y and accumulate */
         if(pt.y <= monitors[SortY[i]].rcMonitor.bottom)
            { /* within current */
             v.y = -monitors[SortY[0]].rcMonitor.top + pt.y;
             found = TRUE;
             break; // we've resolved it
            } /* within current */
        } /* scan Y and accumulate */

     if(!found)
        { /* out of range */
         ::SetLastError(MONITORS_ERR_OUT_OF_BOUNDS);
         return FALSE;  // point is outside virtual screen
        } /* out of range */

     pt = v;
     
     return TRUE;
    } // CMonitors::ScreenToVirtual

/****************************************************************************
*                         CMonitors::VirtualToScreen
* Inputs:
*       CRect & r: IN: Rectangle expressed in virtual coordinates
*                 OUT: Rectangle expressed in screen coordinates
* Result: BOOL
*       TRUE if successful
*       FALSE if error
* Effect: 
*       Converts the coordinates from virtual to screen
****************************************************************************/

BOOL CMonitors::VirtualToScreen(CRect & r)
    {
     if(!VirtualToScreen((CPoint &)r.left))
         return FALSE;
     
     if(!VirtualToScreen((CPoint &)r.right))
        return FALSE;

     return TRUE;
    } // CMonitors::VirtualToScreen

/****************************************************************************
*                         CMonitors::VirtualToScreen
* Inputs:
*       CPOint & pt: IN: Point in virtual coordinates
*                   OUT: Point in screen coordinates
* Result: BOOL
*       TRUE If successful
*       FALSE if error
* Effect: 
*       Converts the coordinates to screen coordinates
* Notes:
*       If the 
****************************************************************************/

BOOL CMonitors::VirtualToScreen(CPoint & pt)
    {
     if(!CheckValid())
        return FALSE;

     if(monitors.GetSize() == 1)
        return TRUE; // Already in virtual coordinates; only one monitor

     MON_TRACE(("VirtualToScreen: pt=(%d,%d)\n", pt.x, pt.y));
     BOOL found = FALSE;
     CPoint s(0,0);

     for(int i = 0; i < VirtualRects.GetSize(); i++)
        { /* scan rects */
         MON_TRACE(("VirtualToScreen: VirtualRects[%d].rcMonitor = (L %d, T %d, R %d, B %d) [%d, %d]\n", 
                    i,
                    VirtualRects[i].left,
                    VirtualRects[i].top,
                    VirtualRects[i].right,
                    VirtualRects[i].bottom,
                    VirtualRects[i].right - VirtualRects[i].left,
                    VirtualRects[i].bottom - VirtualRects[i].top));

         if(VirtualRects[i].PtInRect(pt))
            { /* found it */
             found = TRUE;
             s.x = monitors[i].rcMonitor.left + (pt.x - VirtualRects[i].left);
             s.y = monitors[i].rcMonitor.top  + (pt.y - VirtualRects[i].top);
             break;
            } /* found it */
        } /* scan rects */

     if(found)
        { /* success */
         MON_TRACE(("VirtualToScreen: s=(%d, %d)\n", s.x, s.y));
         pt = s;
         return TRUE;
        } /* success */
     else
        { /* not valid */
         MON_TRACE(("VirtualToScreen: out of bounds\n"));
         ::SetLastError(MONITORS_ERR_OUT_OF_BOUNDS);
         return FALSE;
        } /* not valid */
    } // CMonitors::VirtualToScreen

/****************************************************************************
*                          CMonitors::RightLimitFrom
* Inputs:
*       CPoint & pt: IN: Screen point of origin
*                   OUT: pt.x is now the rightmost point that can be
*                        projected from the original pt.x
* Result: BOOL
*       TRUE if successful
*       FALSE if error
* Effect:
*       Computes the rightmost limit of a line extended from pt.x and which
*       still remains visible, and updates pt.x to represent this point
* GetLastError:
*       MONITORS_ERR_UNINITIALIZED: Either the program failed to call the
*               Initialize method, or the method failed
*       MONITORS_ERR_INIT_FAILURE: This represents an internal error, because
*               it means initialization appeared to work, but the monitors
*               array is empty
*       MONITORS_ERR_OUT_OF_BOUNDS: The input parameter pt is not contained
*               in any monitor
****************************************************************************/

BOOL CMonitors::RightLimitFrom(CPoint & pt)
    {
     //********************************
     // Test all simple cases
     //********************************
     if(!CheckValid())
        return FALSE;

     if(monitors.GetSize() == 1)
        { /* update to width */
         pt.x = monitors[0].rcMonitor.right - 1;
         return TRUE; // Already in virtual coordinates; only one monitor
        } /* update to width */

     //****************************************************************
     // Pass 1: find the monitor that contains pt
     //****************************************************************
     BOOL found = FALSE;
     int first;
     for(first = 0; first < monitors.GetSize(); first++)
        { /* scan monitors */
         CRect r = monitors[SortX[first]].rcMonitor;
         if(r.PtInRect(pt))
            { /* found starting monitor */
             found = TRUE;
             break;
            } /* found starting monitor */
        } /* scan monitors */

     // If no monitor contained the point, we can't do the computation
     if(!found)
        { /* can't find start monitor */
         MON_TRACE(("RightLimitFrom: out of bounds\n"));
         ::SetLastError(MONITORS_ERR_OUT_OF_BOUNDS);
         return FALSE;
        } /* can't find start monitor */

     CPoint result = pt;
     result.x = monitors[SortX[first]].rcMonitor.right;

     // now project outwards
     for(int i = first + 1; i < monitors.GetSize(); i++)
        { /* check end */
         CPoint test = result;
         test.x++; // see if it is in the next monitor
         CRect target = monitors[SortX[i]].rcMonitor;
         if(!target.PtInRect(test))
            break; // we have the max
         result.x = target.right;
        } /* check end */
     pt = result;
     return TRUE;
    } // CMonitors::RightLimitFrom

/****************************************************************************
*                          CMonitors::LeftLimitFrom
* Inputs:
*       CPoint & pt: IN: Starting poin
*                   OUT: Point with .x adjusted to be the maximum left extent
* Result: BOOL
*       TRUE if successful
*       FALSE if error
* Effect: 
*       Computes the leftmost limit of a line extended from pt.x and which
*       still remains visible, and updates pt.x to represent this point
* GetLastError:
*       MONITORS_ERR_UNINITIALIZED: Either the program failed to call the
*               Initialize method, or the method failed
*       MONITORS_ERR_INIT_FAILURE: This represents an internal error, because
*               it means initialization appeared to work, but the monitors
*               array is empty
*       MONITORS_ERR_OUT_OF_BOUNDS: The input parameter pt is not contained
*               in any monitor
****************************************************************************/

BOOL CMonitors::LeftLimitFrom(CPoint & pt)
    {
     if(!CheckValid())
        return FALSE;
     
     if(monitors.GetSize() == 1)
        { /* update to width */
         pt.x = monitors[0].rcMonitor.left;
         return TRUE; // Already in virtual coordinates; only one monitor
        } /* update to width */

     BOOL found = FALSE;
     INT_PTR first;
     for(first = monitors.GetSize() - 1; first >= 0; first--)
        { /* scan monitors */
         CRect r = monitors[SortX[first]].rcMonitor;
         if(r.PtInRect(pt))
            { /* found starting monitor */
             found = TRUE;
             break;
            } /* found starting monitor */
        } /* scan monitors */

     if(!found)
        { /* can't find start monitor */
         MON_TRACE(("LeftLimitFrom: out of bounds\n"));
         ::SetLastError(MONITORS_ERR_OUT_OF_BOUNDS);
         return FALSE;
        } /* can't find start monitor */

     CPoint result = pt;
     result.x = monitors[SortX[first]].rcMonitor.left;

     // now project outwards
     for(INT_PTR i = first - 1; i >= 0; i--)
        { /* check end */
         CPoint test = result;
         test.x--; // see if it is in the next monitor
         CRect target = monitors[SortX[i]].rcMonitor;
         if(!target.PtInRect(test))
            break; // we have the max
         result.x = target.left;
        } /* check end */
     pt = result;
     return TRUE;
    } // CMonitors::LeftLimitFrom

/****************************************************************************
*                           CMonitors::TopLimitFrom
* Inputs:
*       CPoint & pt: IN: Starting point
*                   OUT: Point with .y adjusted to indicate the highest (that is,
*                       lowest-value) point that can be reached contiguously from
*                       the current monitor
* Result: BOOL
*       TRUE if successful
*       FALSE if there is an error
* Effect: 
*       Computes the topmost limit of a line extended from pt.y and which
*       still remains visible, and updates pt.y to represent this point
* GetLastError:
*       MONITORS_ERR_UNINITIALIZED: Either the program failed to call the
*               Initialize method, or the method failed
*       MONITORS_ERR_INIT_FAILURE: This represents an internal error, because
*               it means initialization appeared to work, but the monitors
*               array is empty
*       MONITORS_ERR_OUT_OF_BOUNDS: The input parameter pt is not contained
*               in any monitor
****************************************************************************/

BOOL CMonitors::TopLimitFrom(CPoint & pt)
    {
     if(!CheckValid())
        return FALSE;

     if(monitors.GetSize() == 1)
        { /* update to width */
         pt.y = monitors[0].rcMonitor.top;
         return TRUE; // Already in virtual coordinates; only one monitor
        } /* update to width */

     BOOL found = FALSE;
     INT_PTR first;
     for(first = monitors.GetSize() - 1; first >= 0; first--)
        { /* scan monitors */
         CRect r = monitors[SortY[first]].rcMonitor;
         if(r.PtInRect(pt))
            { /* found starting monitor */
             found = TRUE;
             break;
            } /* found starting monitor */
        } /* scan monitors */

     if(!found)
        { /* can't find start monitor */
         MON_TRACE(("TopLimitFrom: out of bounds\n"));
         ::SetLastError(MONITORS_ERR_OUT_OF_BOUNDS);
         return FALSE;
        } /* can't find start monitor */

     CPoint result = pt;
     result.y = monitors[SortY[first]].rcMonitor.top;

     // now project outwards
     for(INT_PTR i = first - 1; i >= 0; i--)
        { /* check end */
         CPoint test = result;
         test.y--; // see if it is in the next monitor
         CRect target = monitors[SortY[i]].rcMonitor;
         if(!target.PtInRect(test))
            break; // we have the max
         result.y = target.top;
        } /* check end */
     pt = result;
     return TRUE;
    } // CMonitors::TopLimitFrom

/****************************************************************************
*                         CMonitors::BottomLimitFrom
* Inputs:
*       CPoint & pt: IN: A starting point
*                   OUT: The point with the y-coordinate indicating the lowest
*                       (highest-value) y-coordinate reachable from the point
* Result: BOOL
*       TRUE if successful
*       FALSE if any error
* Effect: 
*       Computes the bottommost limit of a line extended from pt.y and which
*       still remains visible, and updates pt.y to represent this point
*       Updates pt.y
* GetLastError:
*       MONITORS_ERR_UNINITIALIZED: Either the program failed to call the
*               Initialize method, or the method failed
*       MONITORS_ERR_INIT_FAILURE: This represents an internal error, because
*               it means initialization appeared to work, but the monitors
*               array is empty
*       MONITORS_ERR_OUT_OF_BOUNDS: The input parameter pt is not contained
*               in any monitor
****************************************************************************/

BOOL CMonitors::BottomLimitFrom(CPoint & pt)
    {
     if(!CheckValid())
        return FALSE;

     if(monitors.GetSize() == 1)
        { /* update to width */
         pt.y = monitors[0].rcMonitor.bottom - 1;
         return TRUE; // Already in virtual coordinates; only one monitor
        } /* update to width */

     BOOL found = FALSE;
     int first;
     for(first = 0; first < monitors.GetSize(); first++)
        { /* scan monitors */
         CRect r = monitors[SortY[first]].rcMonitor;
         if(r.PtInRect(pt))
            { /* found starting monitor */
             found = TRUE;
             break;
            } /* found starting monitor */
        } /* scan monitors */

     if(!found)
        { /* can't find start monitor */
         MON_TRACE(("BottomLimitFrom: out of bounds\n"));
         ::SetLastError(MONITORS_ERR_OUT_OF_BOUNDS);
         return FALSE;
        } /* can't find start monitor */

     CPoint result = pt;
     result.y = monitors[SortY[first]].rcMonitor.bottom;

     // now project outwards
     for(int i = first + 1; i < monitors.GetSize(); i++)
        { /* check end */
         CPoint test = result;
         test.y++; // see if it is in the next monitor
         CRect target = monitors[SortY[i]].rcMonitor;
         if(!target.PtInRect(test))
            break; // we have the max
         result.y = target.bottom;
        } /* check end */
     pt = result;
     return TRUE;
    } // CMonitors::BottomLimitFrom
